/*
    libtree : C++ lib that helps to work with tree-like data structures
    Copyright (C) 2019-2021 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    "json.h"
#include    <fstream>

#define     ARRAY_ELEMENT_NOT_LAST  0
#define     ARRAY_ELEMENT_LAST      1

namespace tree{

//-----------------------------------------------------------------------------
// Push given value into given vector of tagged values

void SaveValue(const string &val, TTaggedValues &dest, const int cell,
        const int array_el){

    if(array_el == ARRAY_ELEMENT_NOT_LAST ||
            array_el == ARRAY_ELEMENT_LAST){

        #ifdef LTREE_DEBUGGING
        cout << "\t- This is an array element" << endl;
        #endif

        dest.PushListEl(val);

        if(array_el == ARRAY_ELEMENT_LAST){
            dest.CloseList(cell);
            }
        }
    else{
        dest.push_back(val, cell);
        }
    }

//-----------------------------------------------------------------------------
// Extract data from needed nodes of a document
// (a special part of the "TTree::Select" method)

bool TJsonTree::SpecialPartOFSelect(const string &file_path,
            vector<TNode> &subtree, TSelectedValues &selected_values,
            TSelectedDataEntry &entry, string &err){

    try{
        ifstream ifs(file_path);
        Json::Reader reader;
        Json::Value root_node;

        // Parsing the file with the use of external lib
        if(!reader.parse(ifs, root_node)){
            const string err_info = reader.getFormattedErrorMessages();
            err = "Can't parse " + file_path + " JSON (" + err_info + ")";
            return false;
            }

        // Selecting values of every needed element of the DOM and putting them
        // into selected_values
        return GetValues(n_branch_drafts, root_node, subtree, selected_values, entry);
        }
    catch(const exception &e){
        err = e.what();
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Get values of those nodes of parsed JSON document that are specified
// in this tree

bool TJsonTree::GetValues(const int branches_q, const Json::Value &current_node,
        vector<TNode> &subtree, TSelectedValues &selected_values,
        TSelectedDataEntry &entry, int level, const int array_el){

    #ifdef LTREE_DEBUGGING
    PrintNewStage(level);
    #endif

    string selected_value;
    bool got_value = false;

    // Different actions for different type of nodes
    switch(current_node.type()){

        case Json::objectValue:

            if(current_node.empty()){
                #ifdef LTREE_DEBUGGING
                cout << "\t- Empty object" << endl;
                #endif
                break;
                }

            // Only a JSON object can form a "level", not any other types of
            // JSON values
            level++;

            #ifdef LTREE_DEBUGGING
            cout << "\tEntering an object" << endl;
            #endif

            UpdateCell(level);

            for(const string &key : current_node.getMemberNames()){

                #ifdef LTREE_DEBUGGING
                cout << "\t- Current element is '" << key << "'" << endl;
                #endif

                // Adding current node to the subtree
                if(PushNodeToSubTree(subtree, key)){

                    // Going into this object
                    if(!GetValues(branches_q, current_node[key], subtree,
                            selected_values, entry, level)){
                        return false;
                        }

                    // Current node either contains a value which is saved
                    // into the selected_values var or is a dead end
                    PopLastElOfSubtree(subtree);
                    }

                #ifdef LTREE_DEBUGGING
                cout << "\tGoing for the next key if there is one... "
                     "(current level is " << level << ", "
                     "cell is " << cell << ", "
                     "cells path is " << CellsPathStr(level) << ")" << endl;
                #endif
                }

            #ifdef LTREE_DEBUGGING
            cout << "\t- There are no more keys left" << endl;
            #endif

            break;

        case Json::arrayValue:

            if(current_node.empty()){
                #ifdef LTREE_DEBUGGING
                cout << "\t- Empty array" << endl;
                #endif
                got_value = true;
                break;
                }

            #ifdef LTREE_DEBUGGING
            cout << "\tEntering an array" << endl;
            #endif

            for(Json::ArrayIndex i = 0; i < current_node.size(); i++){

                level0 = level;

                const int array_el_type = (i == (current_node.size() - 1)) ?
                    ARRAY_ELEMENT_LAST : ARRAY_ELEMENT_NOT_LAST;

                // Going into this array
                if(!GetValues(branches_q, current_node[i], subtree,
                        selected_values, entry, level, array_el_type)){
                    return false;
                    }

                #ifdef LTREE_DEBUGGING
                cout << "\tDo we need to push entry? Current level is "
                     << level << ", cell is " << cell << endl;
                #endif

                // Resetting the counter when leaving this function
                // for a root element
                if(level == 0){
                    #ifdef LTREE_DEBUGGING
                    cout << "\t- Yes" << endl;
                    #endif
                    FinaliseCurrentPass(branches_q, selected_values, entry);
                    }
                #ifdef LTREE_DEBUGGING
                else{
                    cout << "\t- No" << endl;
                    }
                #endif
                }

            break;

        // Rest of the cases are various types of values

        case Json::booleanValue:

            selected_value = (current_node.asBool() ? "1" : "0");

            #ifdef LTREE_DEBUGGING
            cout << "\t- A value, converting to string '"
                 << selected_value << "'" << endl;
            #endif

            got_value = true;

            break;

        case Json::intValue:
        case Json::uintValue:
        case Json::realValue:
        case Json::stringValue:

            selected_value = current_node.asString();

            #ifdef LTREE_DEBUGGING
            cout << "\t- A value, converting to string '"
                 << selected_value << "'" << endl;
            #endif

            got_value = true;

            break;

        case Json::nullValue:

            #ifdef LTREE_DEBUGGING
            cout << "\t- \"Null\" value" << endl;
            #endif

            got_value = true;

            break;

        default:
            cout << "[ERROR: Wrong JSON node type]" << endl;
            return false;
        }

    const int current_cell = CurrentCell(level);

    // If a current cell is not a last new cell then there was a step back in
    // a bypass of a tree and one step should be removed from a path of cells
    if(current_cell != cell && current_node.type() == Json::objectValue){
        set<int> higher_levels;
        for(const auto &[l, _] : cells_path){
            if(l >= level){
                higher_levels.insert(l);
                }
            }
        for(const auto l : higher_levels){
            cells_path.erase(l);
            }
        }

    // Checking formed subtree for intersections with branches of the tree
    const auto [matched_branches, complete_match] = Intersections(subtree);

    // Saving a value if there is a complete match and there is a value
    if(complete_match && got_value){

        // For every branch that completely matched
        for(const auto &matched_branch : matched_branches){

            const int branch_mark = matched_branch.BranchMark();

            if(branch_mark < 0 || branch_mark >= branches_q){
                cout << "[ERROR: branch mark must be in "
                     << "[0; " << branches_q << "), "
                     << "got " << branch_mark << "]" << endl;
                return false;
                }

            const int master_branch = matched_branch.MasterBranch();

            if(matched_branch.IndexEnabled()){
                const int index = UpdateIndex(master_branch);
                selected_value += to_string(index);
                }

            auto &certain_el_of_selection =
                selected_values[master_branch][branch_mark][level];

            SaveValue(selected_value, certain_el_of_selection, current_cell, array_el);

            #ifdef LTREE_DEBUGGING
            PrintAdditionalUpdate(level, master_branch, branch_mark, selected_value);
            selected_values.Print();
            #endif

            // Saving a set of cells that are connected with a current cell
            UpdateCellsConnections(current_cell);
            }
        }

    #ifdef LTREE_DEBUGGING
    cout << "\tLeaving this stage "
         << "(cells path is " << CellsPathStr(level) << ")" << endl;
    #endif

    return true;
    }

//-----------------------------------------------------------------------------

} // namespace tree
