/*
    libtree : C++ lib that helps to work with tree-like data structures
    Copyright (C) 2019-2021 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    "tree.h"
#include    "xml.h"
#include    "json.h"

#ifdef      LTREE_DEBUGGING
#include    "tree_debug.cpp"
#endif

namespace tree{

//-----------------------------------------------------------------------------
// Remove last element from a referenced subtree

void PopLastElOfSubtree(vector<TNode> &subtree){

    if(subtree.empty()){
        return;
        }

    #ifdef LTREE_DEBUGGING
    cout << "\tThrowing away last element of the subtree "
         << "(\"" << subtree.back().Name() << "\")" << endl;
    #endif

    subtree.pop_back();

    #ifdef LTREE_DEBUGGING
    cout << "\tCurrent subtree is ";
    PrintSubTree(subtree, true);
    #endif
    }

//-----------------------------------------------------------------------------
// Find node by specified name and parent ID

TNode* TTree::FindNode(const string &node_name, const int parent_id){

    for(auto &n : nodes){
        if(n.Name() == node_name && n.ParentId() == parent_id){
            return &n;
            }
        }

    return NULL;
    }

//-----------------------------------------------------------------------------
// Construct a tree using given base nodes and a branch draft

bool TTree::Construct(const string &root_name,
        const vector<TBranchDraft> &branch_drafts){

    n_branch_drafts = branch_drafts.size();

    TTreeConstructor tc;

    const auto [data, ok] = tc.Construct(root_name, branch_drafts, Type());
    tie(nodes, branches) = data;

    return ok;
    }

//-----------------------------------------------------------------------------
// Save values of the additional branches if there are such

void TTree::AddItemsOfAdditionalBranches(
        const map<int, vector<string>> &values_of_additional_branches,
        const int master_branch, TSelectedDataEntry &entry,
        const size_t position_in_item){

    #ifdef LTREE_DEBUGGING
    cout << "\tAdding items of additional branches of master branch "
         << master_branch << endl;
    if(values_of_additional_branches.empty()){
        cout << "\t- Empty map!" << endl;
        }
    #endif

    // For all additional branches
    for(const auto &[branch_mark, values] : values_of_additional_branches){

        entry[branch_mark].resize(position_in_item + 1);
        string value_of_item;

        for(const auto &v : values){
            value_of_item += v + ",";
            }

        if(!value_of_item.empty()){
            value_of_item.pop_back();
            }

        entry[branch_mark][position_in_item] = value_of_item;

        #ifdef LTREE_DEBUGGING
        PrintEntryUpdate(branch_mark, entry, 2);
        #endif
        }
    }

//-----------------------------------------------------------------------------
// Push a new entry to a selected data (that is a vector of selected entries)

bool TTree::AddEntry(const int branches_q, TSelectedDataEntry &entry){

    // If entry contains at least one item
    if(first_entry_update){
        return false;
        }

    #ifdef LTREE_DEBUGGING
    cout << "\tAdding new entry:" << endl
         << "\t\t" << entry.Stringify() << endl;
    #endif

    PushBack(selected_data, entry, SELDATA_MEM_STEP);
    entry.Reset(branches_q);
    first_entry_update = true;

    return true;
    }

//-----------------------------------------------------------------------------
// Push given node into given subtree

bool TTree::PushNodeToSubTree(vector<TNode> &subtree, const string &node_name){

    TNode *identified_node = NULL;

    #ifdef LTREE_DEBUGGING
    cout << "\tPushing node to current subtree ";
    PrintSubTree(subtree, true);
    #endif

    // Handling a root node
    if(subtree.empty()){
        if(nodes.empty() || nodes.front().Name() != node_name){
            #ifdef LTREE_DEBUGGING
            cout << "\t- Won't add node " << node_name << " to subtree - "
                "subtree is empty and current node can't be root" << endl;
            #endif
            return false;
            }
        #ifdef LTREE_DEBUGGING
        cout << "\tThis is a root node" << endl;
        #endif
        identified_node = &nodes.front();
        }
    // Handling a non-root node
    else{
        #ifdef LTREE_DEBUGGING
        cout << "\tTrying to find node \"" << node_name << "\" that has "
            "a parent with ID " << subtree.back().Id() << endl;
        #endif
        identified_node = FindNode(node_name, subtree.back().Id());
        }

    if(!identified_node){
        #ifdef LTREE_DEBUGGING
        cout << "\t- Current node cannot be identified" << endl;
        #endif
        return false;
        }

    TNode node(node_name, identified_node->Id(), identified_node->ParentId());
    node.WriteAttribute(identified_node->Attribute());
    PushBack(subtree, node, SUBTREE_MEM_STEP);

    #ifdef LTREE_DEBUGGING
    cout << "\tCurrent subtree is ";
    PrintSubTree(subtree, true);
    #endif

    return true;
    }

//-----------------------------------------------------------------------------
// Search for IDs of branches that are matched with a given subtree, return
// a) an empty vector and a true value if complete match is found or
// b) a vector of partial matches and a false value otherwise.

pair<vector<TBranch>, bool> TTree::Intersections(
        const vector<TNode> &subtree) const{

    vector<TBranch> matched_branches;
    bool complete_match = false;

    // Resolving branch marks
    for(const auto &branch : branches){

        if(subtree.size() > branch.size()){
            continue;
            }

        bool match = true;

        // Check if subtree contents nodes that are not present in the branch
        for(size_t i = 0; i < subtree.size() && i < branch.size(); i++){
            if(subtree.at(i).Name() != branch.at(i).Name()){
                match = false;
                break;
                }
            }

        if(match){
            // Full match
            if(subtree.size() == branch.size()){
                // From now on saving only complete matches
                if(!complete_match){
                    matched_branches.clear();
                    }
                matched_branches.push_back(branch);
                complete_match = true;
                }
            // Is a subtree
            else if(!complete_match){
                matched_branches.push_back(branch);
                }
            }
        }

    #ifdef LTREE_DEBUGGING
    PrintMatches(matched_branches, complete_match);
    #endif

    return {matched_branches, complete_match};
    }

//-----------------------------------------------------------------------------
// Increment current cell number and update a current path of cells if
// the level is changed

void TTree::UpdateCell(const int level){

    if(level != level0){
        cell++;
        cells_path[level] = cell;
        }

    #ifdef LTREE_DEBUGGING
    cout << "\tThis is level " << level << " (prev level is " << level0 << "), "
         << (level != level0 ? "new " : "known ") << "cell " << cell  << ", "
         << "cells path is " << CellsPathStr(level) << endl;
    #endif

    level0 = level;
    }

//-----------------------------------------------------------------------------
// Update a selection index for specified branch

int TTree::UpdateIndex(const int master_branch){

    auto it = selection_index.find(master_branch);
    if(it == selection_index.end()){
        selection_index[master_branch] = 0;
        return 0;
        }

    return ++(it->second);
    }

//-----------------------------------------------------------------------------
// Extract data from those nodes of the document that suit this tree

bool TTree::Select(const string &file_path, string &err){

    #ifdef LTREE_DEBUGGING
    cout << "Starting data selection" << endl;
    #endif

    selected_data.clear();
    err.clear();

    if(nodes.empty()){
        err = "Vector of needed nodes of the document is empty";
        return false;
        }

    // Branch of currently processed nodes of XML DOM
    vector<TNode> subtree;
    // Dict used to store values of additional branches
    TSelectedValues selected_values;
    // Mass of items of the entry
    TSelectedDataEntry entry(n_branch_drafts);

    // This stage is different between child classes
    if(!SpecialPartOFSelect(file_path, subtree, selected_values, entry, err)){
        err = "Can't extract needed data from parsed document "
            "\"" + file_path + "\"";
        return false;
        }

    #ifdef LTREE_DEBUGGING
    cout << "Selected data:" << endl << selected_data.Stringify();
    #endif

    return true;
    }

//-----------------------------------------------------------------------------
// Get a number of a currently processed cell

const int TTree::CurrentCell(const int level) const{

    // Get a cell number of a specified level from a current path of cells
    auto cell_it = cells_path.find(level);

    // If current path of cells does not have a specified level or does have
    const int current_cell = (cell_it == cells_path.end()) ?
        cell : cell_it->second;

    #ifdef LTREE_DEBUGGING
    cout << "\tCurrent cell is " << current_cell << endl;
    #endif

    return current_cell;
    }

//-----------------------------------------------------------------------------
// Update a set of cells that are connected with a current cell

void TTree::UpdateCellsConnections(const int current_cell){

    for(const auto &[_, c] : cells_path){
        // There is no need to record that a cell is connected with itself
        if(c != current_cell){
            cells_connections[current_cell].insert(c);
            }
        }

    #ifdef LTREE_DEBUGGING
    PrintCellsConnections();
    #endif
    }

//-----------------------------------------------------------------------------
// Write selected values into the entry and prepare for a next pass

void TTree::FinaliseCurrentPass(const int branches_q,
        TSelectedValues &selected_values, TSelectedDataEntry &entry){

    const auto it = selected_values.find(-1);

    #ifdef LTREE_DEBUGGING
    cout << "\tGone back to the root" << endl;
    PrintCellsConnections();
    #endif

    // If there is a selection for master branches
    if(it != selected_values.end()){

        const auto selection_for_master_branches = it->second;

        for(const auto &[branch_mark, levels_and_rest] : selection_for_master_branches){
            for(const auto &[l, tagged_values] : levels_and_rest){
                for(const auto &[value, c] : tagged_values){

                    auto is_connected = [&](const int cell_x, const int cell_y){
                        auto it = cells_connections.find(cell_x);
                        if(it == cells_connections.end()){
                            return false;
                            }
                        return it->second.find(cell_y) != it->second.end();
                        };

                    auto condition = [&](const int c_in){
                        return c == c_in || is_connected(c, c_in) || is_connected(c_in, c);
                        };

                    // Saving values of a master branch
                    const size_t position_of_new_item = entry.at(branch_mark).size();
                    PushBack(entry[branch_mark], value, ENTRY_MEM_STEP);

                    #ifdef LTREE_DEBUGGING
                    PrintEntryUpdate(branch_mark, entry);
                    cout << "\tThis is cell " << c << endl;
                    #endif

                    first_entry_update = false;

                    // Saving values of additional branches if there are such
                    const auto values_of_additional_branches =
                        selected_values.Align(branch_mark, condition);

                    AddItemsOfAdditionalBranches(values_of_additional_branches,
                        branch_mark, entry, position_of_new_item);
                    }
                }
            }

        // Adding entry to selected data
        AddEntry(branches_q, entry);
        }

    selected_values.clear();
    selection_index.clear();
    cells_path.clear();
    cells_connections.clear();
    cell = -1;

    #ifdef LTREE_DEBUGGING
    entry_counter = 0;
    #endif
    }

//-----------------------------------------------------------------------------

} // namespace tree
