/*
    libtree : C++ lib that helps to work with tree-like data structures
    Copyright (C) 2019-2021 Alexey Appolonov

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/

#include    "xml.h"

namespace tree{

//-----------------------------------------------------------------------------
// Get a full name (<name_space>:<name>) of a node

string NodeFullName(const xmlpp::Node *node){

    string name;

    if(node){
        name = node->get_name();
        if(!name.empty()){
            string nspace = node->get_namespace_prefix();
            if(!nspace.empty()){
                name = nspace + ":" + name;
                }
            }
        }

    return name;
    }

//-----------------------------------------------------------------------------
// Get value from specified attribute of a given XML node element

void GetValFromSpecifiedAttr(const xmlpp::Element *nodeElement,
        const string &attr_name, string &val){

    const auto attr = nodeElement->get_attribute(attr_name);

    if(attr){
        val = attr->get_value();
        return;
        }

    #ifdef LTREE_DEBUGGING
    cout << "\t- CAN'T RETRIEVE THE VALUE! "
         << "Checking the attributes one by one:" << endl;
    #endif

    for(const auto &a : nodeElement->get_attributes()){

        #ifdef LTREE_DEBUGGING
        cout << "\t\t" << a->get_name() << " = " << a->get_value() << endl;
        #endif

        if(attr_name == a->get_name()){

            val = a->get_value();

            #ifdef LTREE_DEBUGGING
            cout << "\t\t- Found!" << endl;
            #endif

            break;
            }
        }
    }

//----------------------------------------------------------------------------
// Get value from a content of a text XML node

void GetValFromContent(const xmlpp::TextNode *nodeText, string &val){

    const string &content = nodeText->get_content();

    if(!content.empty()){
        val = content;
        }
    }

//-----------------------------------------------------------------------------
// Extract data from the needed nodes of a document
// (a special part of the "TTree::Select" method)

bool TXmlTree::SpecialPartOFSelect(const string &file_path,
        vector<TNode> &subtree, TSelectedValues &selected_values,
        TSelectedDataEntry &entry, string &err){

    try{
        // Parsing the file with the use of external lib
        xmlpp::DomParser parser(file_path, validate);
        if(!parser){
            err = "Can't parse " + file_path + " XML";
            return false;
            }

        // Get DOM of the XML-document
        xmlpp::Document *dom = parser.get_document();

        // Get root element of the XML-document
        xmlpp::Element *root_node = dom->get_root_node();

        // Selecting values of every needed element of the DOM and putting them
        // into selected_values
        return GetValues(n_branch_drafts, root_node, subtree, selected_values, entry);
        }
    catch(const exception &e){
        err = e.what();
        }

    return false;
    }

//-----------------------------------------------------------------------------
// Get values of those nodes of parsed XML document that are specified
// in this tree

bool TXmlTree::GetValues(const int branches_q, const xmlpp::Node *current_node,
            vector<TNode> &subtree, TSelectedValues &selected_values,
            TSelectedDataEntry &entry, const int level){

    #ifdef LTREE_DEBUGGING
    PrintNewStage(level);
    #endif

    if(!current_node){
        cout << "[ERROR: No current node to follow]" << endl;
        return false;
        }

    const auto nodeContent = dynamic_cast<const xmlpp::ContentNode*>(current_node);
    const auto nodeText = dynamic_cast<const xmlpp::TextNode*>(current_node);
    const auto nodeElement = dynamic_cast<const xmlpp::Element*>(current_node);

    if(nodeText){
        // Skipping empty text nodes
        if(nodeText->is_white_space()){
            #ifdef LTREE_DEBUGGING
            cout << "\t- Empty text node" << endl;
            #endif
            return true;
            }
        }
    // No need to save cell number of a text node
    else{
        UpdateCell(level);
        }

    #ifdef LTREE_DEBUGGING
    cout << "\tName of a current node is \""
         << NodeFullName(current_node) << "\"" << endl;
    #endif

    // Adding current node to the subtree
    if(!PushNodeToSubTree(subtree, NodeFullName(current_node))){
        return true;
        }

    // Checking formed subtree for intersections with branches of the tree
    const auto [matched_branches, complete_match] = Intersections(subtree);

    if(complete_match){

        const int current_cell = CurrentCell(level);

        // For every branch that completely matched
        for(const auto &matched_branch : matched_branches){

            string selected_value;
            const TNode &leaf = matched_branch.back();
            const string &attr_name = leaf.Attribute();

            // If node has some attributes
            if(nodeElement && !attr_name.empty()){

                #ifdef LTREE_DEBUGGING
                cout << "\tThis node has an attribute "
                     << "\"" << attr_name << "\"" << endl;
                #endif

                GetValFromSpecifiedAttr(nodeElement, attr_name, selected_value);
                }

            // If node holds a value in it's content
            else if(nodeText){
                GetValFromContent(nodeText, selected_value);
                }

            // This node does not contain a value
            else{
                continue;
                }

            #ifdef LTREE_DEBUGGING
            cout << "\tProcessing matched branch..." << endl;
            matched_branch.Print();
            #endif

            // There is a value in an attribute or in a content of a node

            const int branch_mark = matched_branch.BranchMark();

            if(branch_mark < 0 || branch_mark >= branches_q){
                cout << "[ERROR: branch mark must be in "
                     << "[0; " << branches_q << "), "
                     << "got " << branch_mark << "]" << endl;
                return false;
                }

            const int master_branch = matched_branch.MasterBranch();

            if(matched_branch.IndexEnabled()){
                const int index = UpdateIndex(master_branch);
                selected_value += to_string(index);
                }

            auto &certain_el_of_selection =
                selected_values[master_branch][branch_mark][level];

            certain_el_of_selection.push_back(selected_value, current_cell);

            #ifdef LTREE_DEBUGGING
            PrintAdditionalUpdate(level, master_branch, branch_mark, selected_value);
            selected_values.Print();
            #endif

            // Updating a set of cells that are connected with a current cell
            UpdateCellsConnections(current_cell);
            }
        }

    // If a current node is not a content node
    if(!nodeContent){

        // For all childs of this node
        for(const auto &child : current_node->get_children()){

            // Go deeper in the XML DOM tree
            if(!GetValues(branches_q, child, subtree, selected_values, entry,
                    level + 1)){
                return false;
                }
            }
        }

    // Current node has no more children nodes to enter
    PopLastElOfSubtree(subtree);

    auto it = cells_path.find(level);
    if(it != cells_path.end()){
        cells_path.erase(it);
        }

    if(level == 1){
        FinaliseCurrentPass(branches_q, selected_values, entry);
        }

    return true;
    }

//-----------------------------------------------------------------------------

} // namespace tree
